# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/2A. Speaker Embeddings.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/2A. Speaker Embeddings.ipynb 3
import os
import io
import time
import torch
import torchaudio

# %% ../nbs/2A. Speaker Embeddings.ipynb 4
from pathlib import Path
import json
from fastprogress import progress_bar, master_bar
import numpy as np
import random

from torch import nn
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader

from fastcore.script import *

from . import vad, utils, wh_transcribe, vq_stoks
import webdataset as wds

from speechbrain.pretrained import EncoderClassifier

# %% ../nbs/2A. Speaker Embeddings.ipynb 7
def calc_len(x):
    x['seconds'] = torch.tensor(x['tend'] - x['tstart'])
    return x

def chunked_dataset(input, key, bs=16):
    p = Path(input)
    rename_files = vad.fix_dots_in_names if 'librilight' in input else None
    ds = vad.load_dataset(input, rename_files=rename_files).compose(
        wh_transcribe.merge_in(wds.WebDataset(str(p.parent/vad.flac_to_vad_name(input, key))).decode()),
        lambda x: wh_transcribe.split_to_chunks(x, pad_to_seconds=30),
        utils.resampler(16000, 'samples_16k'),
        wds.map(calc_len),
        wds.to_tuple('__key__', 'samples_16k', 'seconds'),
        wds.shuffle(6000),
        wds.batched(bs),
    )
    dl = DataLoader(ds, num_workers=1, batch_size=None)
    return dl

# %% ../nbs/2A. Speaker Embeddings.ipynb 15
def flac_to_spk_emb_name(input, key='raw'):
    return input.rsplit("/", 1)[1].replace(key, f'spk_emb') + ".gz"

@call_parse
def process_shard(
    input:str,          # input shard URL/path
    output:str=None,    # output shard URL/path
    bs:int=None,        # batch size (16 uses around 11GB of VRAM)
    n_samples:int=None, # limit the number of samples (useful for quick benchmarking)
    key:str='flac',
):
    if output is None: output = flac_to_spk_emb_name(input, key)
    if bs is None: bs = 16
    if n_samples is None: n_samples = 'noinfer'
    else: n_samples = n_samples // bs

    dl = chunked_dataset(input, key, bs)
    
    classifier = EncoderClassifier.from_hparams("speechbrain/spkrec-ecapa-voxceleb",
                                                savedir="~/.cache/speechbrain/",
                                                run_opts={"device": "cuda"})
    
    tmp = output+".tmp"
    with wds.TarWriter(tmp) as sink:
        for keys, tends, tstarts, samples in progress_bar(dl, total=n_samples):
            with torch.no_grad():
                embs = classifier.encode_batch(samples, wav_lens=(torch.tensor(tends)-torch.tensor(tstarts))/30).squeeze(1)
            for key, emb in zip(keys, embs):
                sink.write({
                    "__key__": key,
                    "spk_emb.npy": emb.cpu().numpy(),
                })
    os.rename(tmp, output)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/0. Verify webdataset archives.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/0. Verify webdataset archives.ipynb 2
import webdataset as wds
from . import vad, wh_transcribe, vq_stoks
from pathlib import Path
from fastcore.script import call_parse
from fastprogress import progress_bar

# %% ../nbs/0. Verify webdataset archives.ipynb 7
@call_parse
def process_shard(
    input:str,          # input shard URL/path
    processed:str = None, # processed data path
    length:int = None
):
    if type(input) == str and not input.endswith('.tar'):
        input = [str(x) for x in Path(input).glob('*.tar')]
    ds = vad.load_dataset(input, decode=False)
    if processed:
        ds = ds.compose(
            wds.decode(wds.torch_audio),
            vq_stoks.merge_in(vq_stoks.derived_dataset(processed, 'vad')),
            wds.map_dict(**{"vad.npy":wh_transcribe.chunk_merger}),
            wh_transcribe.split_to_chunks,
            vq_stoks.merge_in(vq_stoks.derived_dataset(processed, 'base.en-txt')),
        )
        if length is None: length = len(input)
        bar = iter(progress_bar(range(length)))
    seen_urls = {}
    for s in wds.WebLoader(ds, num_workers=16, batch_size=None):
        if processed:
#             print(s['__url__'], s['__key__'])
            if s['__url__'] not in seen_urls:
                next(bar)
                seen_urls[s['__url__']] = True
        else:
            if 'flac' not in s or 'json' not in s:
                print(s['__url__'], s.keys())
                break



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Split and shuffle webdataset.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/Split and shuffle webdataset.ipynb 1
import webdataset as wds
import pylab as plt
from pathlib import Path
import shutil
from fastprogress import progress_bar
from fastcore.script import call_parse
import numpy as np
import random
from collections import Counter

# %% ../nbs/Split and shuffle webdataset.ipynb 9
@call_parse
def split_dataset(
    input:str,
    dest_folder:Path,
    val_size=512,
    train_shards=400
):
    dest = Path(dest_folder)
    if dest.exists(): shutil.rmtree(dest)
    dest.mkdir(parents=True)

    if isinstance(input, (Path, str)):
        path = Path(input)
        if path.is_dir():
            glob = '*.tar.gz'
        else:
            glob = path.name
            path = path.parent
        input = Path(path).glob(glob)
    elif isinstance(input, list):
        pass
    else:
        raise ArgumentError("input should be either a list or a path with an optional glob specifier")
    shards = [str(x) for x in input]

    ds = wds.WebDataset([shards[0]])
    N = len([True for _ in ds]) * len(shards)
    
    ds = wds.WebDataset(shards, shardshuffle=True)
    dl = wds.WebLoader(ds, num_workers=32, batch_size=None).shuffle(200000)

    print(f"Writing the datasets ({N} samples):")
    with wds.TarWriter(str(Path(dest_folder)/"val.tar.gz")) as val_sink:
        with wds.ShardWriter(str(Path(dest_folder)/"train-%06d.tar.gz"), maxcount=N//(train_shards+1)) as train_sink:
            for i, s in enumerate(progress_bar(dl, total='noinfer')):
                if i < val_size:
                    val_sink.write(s)
                else:
                    train_sink.write(s)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/0. Download models.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/0. Download models.ipynb 1
from fastcore.script import call_parse
import whisperx
import whisper

# %% ../nbs/0. Download models.ipynb 3
@call_parse
def main():
    whisper.load_model('base.en')
    whisper.load_model('small.en')
    whisperx.vad.load_vad_model('cpu')
    whisperx.asr.load_model('medium.en', "cpu", compute_type="float16", language='en')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/A3. Download Project Guttenberg audiobooks.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/A3. Download Project Guttenberg audiobooks.ipynb 2
import webdataset as wds
from datasets import load_dataset
from pathlib import Path
from fastprogress import progress_bar
import requests
from fastcore.script import call_parse
import torch
import webdataset as wds
import pyarrow
import torchaudio
import tempfile
import os

# %% ../nbs/A3. Download Project Guttenberg audiobooks.ipynb 10
import time

@call_parse
def main(input_dataset:str, # input dataset path
         dest_dataset_name:str = "guttenberg-audiobooks", # the output folder and file name
        ):
    dest_folder = Path(dest_dataset_name)
    dest_folder.mkdir(exist_ok=True)
    
    # this is too slow:
#     ds = load_dataset('TwoAbove/the-project-gutenberg-open-audiobook-collection', streaming=True)
#     print(ds['train'][0])
#     ds = ds.cast_column('mp3', np.object_)

    def load_parquet(shards):
        for sh in shards:
            df = pyarrow.parquet.ParquetFile(sh['url'])
            for batch in df.iter_batches(1):
                s = {k:v[0] for k,v in batch.to_pydict().items()}
                s['__url__'] = sh['url']
                yield s

    def try_decode(samples):
        with tempfile.TemporaryDirectory() as dirname:
            fname = os.path.join(dirname, "file.mp3")
            for s in samples:
                if s['mp3']: # validate mp3 data
                    if s['mp3'][:3] != b"ID3":
                        s['invalid'] = True
                yield s

    ds = wds.DataPipeline(
        wds.SimpleShardList([str(x) for x in Path(input_dataset).glob('*.parquet')]),
        wds.split_by_worker,
        load_parquet,
        try_decode,
    )
    
    with open('missing_urls', 'w') as missing:
        with open('invalid_urls', 'w') as invalid:
            with wds.ShardWriter(str(dest_folder/f"{dest_dataset_name}-raw-%06d.tar"), maxsize=5e9) as sink:
                for x in progress_bar(wds.WebLoader(ds, num_workers=8, batch_size=None), total='noinfer'):
                    if not x['mp3']:
                        missing.write(f"{x['__url__']} {x['link']}\n")
                    elif 'invalid' in x:
                        invalid.write(f"{x['__url__']} {x['link']}\n")
                    else:
                        sink.write({
                            # make sure there are no dots in the key
                            '__key__': Path(x['link']).name.rsplit('.', 1)[0].replace('.', '_'),
                            'title.txt': x['title'],
                            'mp3': x['mp3'],
                        })



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/2C. Whisper semantic embedding extraction.ipynb.

# %% auto 0
__all__ = ['load_model', 'encode_semantic', 'extract_semantic']

# %% ../nbs/2C. Whisper semantic embedding extraction.ipynb 3
import torch
import torchaudio

from pathlib import Path
from fastprogress import progress_bar, master_bar
from fastcore.script import *

import whisper
from .extract_acoustic import load

# %% ../nbs/2C. Whisper semantic embedding extraction.ipynb 7
def load_model():
    return whisper.load_model('tiny.en')

# %% ../nbs/2C. Whisper semantic embedding extraction.ipynb 28
# same as above but rolled into a function
def encode_semantic(whmodel, audio):
    """Encode the given `audio` (tensor or file name) into Whisper embeddings and lists of text tokens.
    Uses the given `whmodel` (see `load_model`).
    """
    if isinstance(audio, (Path, str)):
        audio = load(audio, newsr=whisper.audio.SAMPLE_RATE)
    mel = whisper.log_mel_spectrogram(audio[0,0])
    embs = []
    toks = []
    for start in range(0, mel.shape[-1], whisper.audio.N_FRAMES):
        sample = mel[:,start:]
        with torch.no_grad():
            padded = whisper.audio.pad_or_trim(sample, whisper.audio.N_FRAMES).unsqueeze(0)
            emb = whmodel.encoder(padded)
            tokens = whmodel.decode(emb, whisper.DecodingOptions(language='en', suppress_blank=False, suppress_tokens=False))[0].tokens
            embs.append(emb.cpu())
            toks.append(tokens)
    return torch.stack(embs, axis=0), toks

# %% ../nbs/2C. Whisper semantic embedding extraction.ipynb 30
@call_parse
def extract_semantic(
        srcdir:Path,  # source dir, should contain *.flac files
        outdir:Path,  # output dir, will get the *.semb and *.ttoks files
        layer='last', # the layer to extract the embeddings from
    ): 
    "Convert audio files to .semb files with Whisper embeddings and .ttoks with text tokens"
    model = load_model()
    suffix = '.semb'
    
    if layer != 'last':
        layer = int(layer)
        N = len(model.encoder.blocks)
        for i in range(N,layer,-1):
            print("Removing layer", i)
            del model.encoder.blocks[i-1]
        model.encoder.ln_post = torch.nn.Identity()
        suffix += f'-{layer}'
    
    outdir.mkdir(exist_ok=True, parents=True)
    for name in progress_bar(list(srcdir.rglob('*.flac'))):
        embs, toks = encode_semantic(model, name)
        torch.save(embs, outdir/name.with_suffix(suffix).name)
        torch.save(toks, outdir/name.with_suffix('.ttoks').name)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Model inference tests-Copy2.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/Model inference tests-Copy2.ipynb 1
import os
# os.environ['TORCH_LOGS'] = ""
# os.environ['TORCH_LOGS'] = "+guards,+perf_hints,+dynamo"
# os.environ['TORCHDYNAMO_VERBOSE'] = '1'

# %% ../nbs/Model inference tests-Copy2.ipynb 2
import torch
#import pylab as plt
from fastprogress import progress_bar
import time

#from torch.profiler import profile, record_function, ProfilerActivity

# %% ../nbs/Model inference tests-Copy2.ipynb 4
from whisperspeech.pipeline import Pipeline

# %% ../nbs/Model inference tests-Copy2.ipynb 7
pipe = Pipeline(t2s_ref='t2s-small-en+pl.model', s2a_ref='s2a-q4-tiny-en+pl.model', use_kv_cache=False)

for emb in [pipe.t2s.embeddings.embedding, pipe.t2s.embeddings.embedding]:
    emb.convert_for_eval()
for emb in pipe.s2a.embds.embeddings:
    emb.convert_for_eval()

for l in pipe.s2a.decoder.layers:
    l.setup_kv_cache(1, 2250, 750)
#     l.attn.convert_for_eval()
#     l.cross_attn.convert_for_eval()
# for l in pipe.s2a.encoder:
#     l.attn.convert_for_eval()

# torch._dynamo.reset()
# pipe.t2s.encoder = torch.compile(pipe.t2s.encoder, mode="reduce-overhead", fullgraph=True)
# pipe.t2s.decoder = torch.compile(pipe.t2s.decoder, mode="reduce-overhead", fullgraph=True)

# pipe.s2a.generate_next = torch.compile(pipe.s2a.generate_next, mode="reduce-overhead", fullgraph=True)

# %% ../nbs/Model inference tests-Copy2.ipynb 10
#pipe.vocoder.decode_to_notebook(pipe.generate_atoks("Welcome to Burger Shack!"))
